@@ -2,11 +2,10 @@
 
 .. testsetup::
 
-    from pyroute2 import IPMock as IPRoute
     from pyroute2 import NDB
     from pyroute2 import config
 
-    config.mock_iproute = True
+    config.mock_netlink = True
 
 
 .. testsetup:: preset_1
@@ -14,12 +13,12 @@
     from pyroute2 import NDB
     from pyroute2 import config
 
-    config.mock_iproute = True
+    config.mock_netlink = True
     ndb = NDB(
         sources=[
-            {'target': 'localhost', 'kind': 'IPMock'},
-            {'target': 'worker1.sample.com', 'kind': 'IPMock'},
-            {'target': 'worker2.sample.com', 'kind': 'IPMock'},
+            {'target': 'localhost', 'kind': 'local'},
+            {'target': 'worker1.sample.com', 'kind': 'local'},
+            {'target': 'worker2.sample.com', 'kind': 'local'},
         ]
     )
 
@@ -27,7 +26,7 @@
 
     from pyroute2 import NDB
     from pyroute2 import config
-    config.mock_iproute = True
+    config.mock_netlink = True
     ndb = NDB()
     ndb.interfaces.create(ifname='eth1', kind='dummy').commit()
     ndb.interfaces.create(ifname='br0', kind='bridge').commit()
@@ -37,7 +36,7 @@
 
     from pyroute2 import NDB
     from pyroute2 import config
-    config.mock_iproute = True
+    config.mock_netlink = True
     ndb = NDB()
     ndb.interfaces.create(ifname='br0', kind='bridge').commit()
     ndb.interfaces['br0'].add_port('eth0').commit()
@@ -151,7 +150,7 @@ Bridge and bond ports
 
 Add bridge and bond ports one can use specific API:
 
-.. testcode:: preset_br0_1
+.. code::
 
     with ndb.interfaces['br0'] as br0:
         br0.add_port('eth0')
@@ -165,7 +164,7 @@ Add bridge and bond ports one can use specific API:
 
 To remove a port:
 
-.. testcode:: preset_br0_2
+.. code::
 
     with ndb.interfaces['br0'] as br0:
         br0.del_port('eth0')
@@ -173,7 +172,7 @@ To remove a port:
 Or by setting the master property on a port, in the same
 way as with `IPRoute`:
 
-.. testcode:: preset_br0_1
+.. code::
 
     index = ndb.interfaces['br0']['index']
 
@@ -188,6 +187,7 @@ way as with `IPRoute`:
 
 import errno
 import json
+import os
 import traceback
 
 from pyroute2.common import basestring
@@ -197,20 +197,21 @@ from pyroute2.netlink.rtnl.ifinfmsg import ifinfmsg
 from pyroute2.netlink.rtnl.p2pmsg import p2pmsg
 from pyroute2.requests.link import LinkFieldFilter
 
-from ..auth_manager import AuthManager, check_auth
-from ..objects import RTNL_Object
+from ..objects import AsyncObject, RTNL_Object
+from ..sync_api import Flags, SyncView
 
 
-def load_ifinfmsg(schema, target, event):
+async def load_ifinfmsg(schema, sources, target, event):
     #
     # link goes down: flush all related routes
     #
     if not event['flags'] & 1:
         schema.execute(
-            'DELETE FROM routes WHERE '
-            'f_target = %s AND '
-            'f_RTA_OIF = %s OR f_RTA_IIF = %s'
-            % (schema.plch, schema.plch, schema.plch),
+            '''
+               DELETE FROM routes WHERE
+               f_target = ? AND
+               f_RTA_OIF = ? OR f_RTA_IIF = ?
+            ''',
             (target, event['index'], event['index']),
         )
     #
@@ -232,7 +233,7 @@ def load_ifinfmsg(schema, target, event):
     #
     if event['family'] == AF_BRIDGE:
         #
-        schema.load_netlink('af_bridge_ifs', target, event)
+        await schema.load_netlink('af_bridge_ifs', sources, target, event)
         try:
             vlans = event.get_attr('IFLA_AF_SPEC').get_attrs(
                 'IFLA_BRIDGE_VLAN_INFO'
@@ -245,22 +246,21 @@ def load_ifinfmsg(schema, target, event):
         # flush the old vlans info
         schema.execute(
             '''
-                       DELETE FROM af_bridge_vlans
-                       WHERE
-                           f_target = %s
-                           AND f_index = %s
-                       '''
-            % (schema.plch, schema.plch),
+               DELETE FROM af_bridge_vlans
+               WHERE
+                   f_target = ?
+                   AND f_index = ?
+            ''',
             (target, event['index']),
         )
         for v in vlans:
             v['index'] = event['index']
             v['header'] = {'type': event['header']['type']}
-            schema.load_netlink('af_bridge_vlans', target, v)
+            await schema.load_netlink('af_bridge_vlans', sources, target, v)
 
         return
 
-    schema.load_netlink('interfaces', target, event)
+    await schema.load_netlink('interfaces', sources, target, event)
     #
     # load ifinfo, if exists
     #
@@ -268,7 +268,7 @@ def load_ifinfmsg(schema, target, event):
         linkinfo = event.get_attr('IFLA_LINKINFO')
         if linkinfo is not None:
             iftype = linkinfo.get_attr('IFLA_INFO_KIND')
-            table = 'ifinfo_%s' % iftype
+            table = f'ifinfo_{iftype}'
             if iftype == 'gre':
                 ifdata = linkinfo.get_attr('IFLA_INFO_DATA')
                 local = ifdata.get_attr('IFLA_GRE_LOCAL')
@@ -277,7 +277,7 @@ def load_ifinfmsg(schema, target, event):
                 p2p['index'] = event['index']
                 p2p['family'] = 2
                 p2p['attrs'] = [('P2P_LOCAL', local), ('P2P_REMOTE', remote)]
-                schema.load_netlink('p2p', target, p2p)
+                await schema.load_netlink('p2p', sources, target, p2p)
             elif iftype == 'veth':
                 link = event.get_attr('IFLA_LINK')
                 ifname = event.get_attr('IFLA_IFNAME')
@@ -286,14 +286,14 @@ def load_ifinfmsg(schema, target, event):
                 if (not link) and (
                     (target,) in schema.fetch('SELECT f_target FROM SOURCES')
                 ):
-                    schema.log.debug('reload veth %s' % event['index'])
+                    schema.log.debug(f'reload veth {event["index"]}')
                     try:
-                        update = schema.sources[target].api(
+                        update = await sources[target].api(
                             'link', 'get', index=event['index']
                         )
                         update = tuple(update)[0]
-                        return schema.load_netlink(
-                            'interfaces', target, update
+                        return await schema.load_netlink(
+                            'interfaces', sources, target, update
                         )
                     except NetlinkError as e:
                         if e.code == errno.ENODEV:
@@ -304,7 +304,7 @@ def load_ifinfmsg(schema, target, event):
                 if ifdata is not None:
                     ifdata['header'] = {}
                     ifdata['index'] = event['index']
-                    schema.load_netlink(table, target, ifdata)
+                    await schema.load_netlink(table, sources, target, ifdata)
 
 
 ip_tunnels = ('gre', 'gretap', 'ip6gre', 'ip6gretap', 'ip6tnl', 'sit', 'ipip')
@@ -386,7 +386,7 @@ supported_ifinfo = {x: ifinfmsg.ifinfo.data_map[x] for x in ifinfo_names}
 # load supported ifinfo
 #
 for name, data in supported_ifinfo.items():
-    name = 'ifinfo_%s' % name
+    name = f'ifinfo_{name}'
     init['classes'].append([name, data])
     schema = (
         data.sql_schema()
@@ -410,7 +410,7 @@ def _cmp_master(self, value):
     return False
 
 
-class Vlan(RTNL_Object):
+class Vlan(AsyncObject):
     table = 'af_bridge_vlans'
     msg_class = ifinfmsg.af_spec_bridge.vlan_info
     api = 'vlan_filter'
@@ -418,28 +418,19 @@ class Vlan(RTNL_Object):
     @classmethod
     def _count(cls, view):
         if view.chain:
-            return view.ndb.task_manager.db_fetchone(
-                'SELECT count(*) FROM %s WHERE f_index = %s'
-                % (view.table, view.ndb.schema.plch),
+            return view.ndb.schema.fetchone(
+                f'SELECT count(*) FROM {view.table} WHERE f_index = ?',
                 [view.chain['index']],
             )
         else:
-            return view.ndb.task_manager.db_fetchone(
-                'SELECT count(*) FROM %s' % view.table
+            return view.ndb.schema.fetchone(
+                f'SELECT count(*) FROM {view.table}'
             )
 
     @classmethod
     def _dump_where(cls, view):
         if view.chain:
-            plch = view.ndb.schema.plch
-            where = '''
-                    WHERE
-                        main.f_target = %s AND
-                        main.f_index = %s
-                    ''' % (
-                plch,
-                plch,
-            )
+            where = 'WHERE main.f_target = ? AND main.f_index = ?'
             values = [view.chain['target'], view.chain['index']]
         else:
             where = ''
@@ -462,7 +453,7 @@ class Vlan(RTNL_Object):
               '''
         yield ('target', 'tflags', 'vid', 'ifname')
         where, values = cls._dump_where(view)
-        for record in view.ndb.task_manager.db_fetch(req + where, values):
+        for record in view.ndb.schema.fetch(req + where, values):
             yield record
 
     @staticmethod
@@ -472,14 +463,6 @@ class Vlan(RTNL_Object):
 
     def __init__(self, *argv, **kwarg):
         kwarg['iclass'] = ifinfmsg.af_spec_bridge.vlan_info
-        if 'auth_managers' not in kwarg or kwarg['auth_managers'] is None:
-            kwarg['auth_managers'] = []
-        log = argv[0].ndb.log.channel('vlan auth')
-        kwarg['auth_managers'].append(
-            AuthManager(
-                {'obj:read': True, 'obj:list': True, 'obj:modify': False}, log
-            )
-        )
         super(Vlan, self).__init__(*argv, **kwarg)
 
     def make_req(self, prime):
@@ -495,7 +478,7 @@ class Vlan(RTNL_Object):
         return self.make_req(prime)
 
 
-class Interface(RTNL_Object):
+class Interface(AsyncObject):
     table = 'interfaces'
     msg_class = ifinfmsg
     api = 'link'
@@ -506,32 +489,24 @@ class Interface(RTNL_Object):
         'alt_ifname_list': lambda x: list(json.loads(x or '[]'))
     }
     field_filter = LinkFieldFilter
+    old_ifname = None
 
     @classmethod
     def _count(cls, view):
         if view.chain:
-            return view.ndb.task_manager.db_fetchone(
-                'SELECT count(*) FROM %s WHERE f_IFLA_MASTER = %s'
-                % (view.table, view.ndb.schema.plch),
+            return view.ndb.schema.fetchone(
+                f'SELECT count(*) FROM {view.table} WHERE f_IFLA_MASTER = ?',
                 [view.chain['index']],
             )
         else:
-            return view.ndb.task_manager.db_fetchone(
-                'SELECT count(*) FROM %s' % view.table
+            return view.ndb.schema.fetchone(
+                f'SELECT count(*) FROM {view.table}'
             )
 
     @classmethod
     def _dump_where(cls, view):
         if view.chain:
-            plch = view.ndb.schema.plch
-            where = '''
-                    WHERE
-                        f_target = %s AND
-                        f_IFLA_MASTER = %s
-                    ''' % (
-                plch,
-                plch,
-            )
+            where = 'WHERE f_target = ? AND f_IFLA_MASTER = ?'
             values = [view.chain['target'], view.chain['index']]
         else:
             where = 'WHERE f_index != 0'
@@ -558,18 +533,16 @@ class Interface(RTNL_Object):
             'kind',
         )
         where, values = cls._dump_where(view)
-        for record in view.ndb.task_manager.db_fetch(req + where, values):
+        for record in view.ndb.schema.fetch(req + where, values):
             yield record
 
     def mark_tflags(self, mark):
-        plch = (self.schema.plch,) * 3
         self.schema.execute(
             '''
-                            UPDATE interfaces SET
-                                f_tflags = %s
-                            WHERE f_index = %s AND f_target = %s
-                            '''
-            % plch,
+               UPDATE interfaces SET
+                   f_tflags = ?
+               WHERE f_index = ? AND f_target = ?
+            ''',
             (mark, self['index'], self['target']),
         )
 
@@ -641,11 +614,10 @@ class Interface(RTNL_Object):
         if isinstance(right, basestring):
             return right == left['ifname'] or right == left['address']
 
-    @check_auth('obj:modify')
     def add_vlan(self, spec):
         def do_add_vlan(self, mode, spec):
             try:
-                method = getattr(self.vlan.create(spec), mode)
+                method = getattr(self.vlans.create(spec), mode)
                 return [method()]
             except Exception as e_s:
                 e_s.trace = traceback.format_stack()
@@ -654,11 +626,10 @@ class Interface(RTNL_Object):
         self._apply_script.append((do_add_vlan, {'spec': spec}))
         return self
 
-    @check_auth('obj:modify')
     def ensure_vlan(self, spec):
         def do_ensure_vlan(self, mode, spec):
             try:
-                method = getattr(self.vlan.create(spec), mode)
+                method = getattr(self.vlans.create(spec), mode)
                 return [method()]
             except KeyError:
                 return []
@@ -669,11 +640,10 @@ class Interface(RTNL_Object):
         self._apply_script.append((do_ensure_vlan, {'spec': spec}))
         return self
 
-    @check_auth('obj:modify')
     def del_vlan(self, spec):
         def do_del_vlan(self, mode, spec):
             try:
-                method = getattr(self.vlan[spec].remove(), mode)
+                method = getattr(self.vlans[spec].remove(), mode)
                 return [method()]
             except Exception as e_s:
                 e_s.trace = traceback.format_stack()
@@ -682,7 +652,6 @@ class Interface(RTNL_Object):
         self._apply_script.append((do_del_vlan, {'spec': spec}))
         return self
 
-    @check_auth('obj:modify')
     def add_neighbour(self, spec=None, **kwarg):
         spec = spec or kwarg
 
@@ -697,7 +666,6 @@ class Interface(RTNL_Object):
         self._apply_script.append((do_add_neighbour, {'spec': spec}))
         return self
 
-    @check_auth('obj:modify')
     def ensure_neighbour(self, spec=None, **kwarg):
         spec = spec or kwarg
 
@@ -714,7 +682,6 @@ class Interface(RTNL_Object):
         self._apply_script.append((do_ensure_neighbour, {'spec': spec}))
         return self
 
-    @check_auth('obj:modify')
     def del_neighbour(self, spec=None, **kwarg):
         spec = spec or dict(kwarg)
 
@@ -723,11 +690,13 @@ class Interface(RTNL_Object):
             if isinstance(spec, basestring):
                 specs = [spec]
             elif callable(spec):
-                specs = self.ipaddr.dump()
+                specs = self.neighbours.dump()
                 specs.select_records(spec)
+                specs.materialize()
             else:
-                specs = self.ipaddr.dump()
+                specs = self.neighbours.dump()
                 specs.select_records(**spec)
+                specs.materialize()
             for sp in specs:
                 try:
                     method = getattr(self.neighbours.locate(sp).remove(), mode)
@@ -744,7 +713,6 @@ class Interface(RTNL_Object):
         self._apply_script.append((do_del_neighbour, {'spec': spec}))
         return self
 
-    @check_auth('obj:modify')
     def add_ip(self, spec=None, **kwarg):
         spec = spec or kwarg
 
@@ -759,7 +727,6 @@ class Interface(RTNL_Object):
         self._apply_script.append((do_add_ip, {'spec': spec}))
         return self
 
-    @check_auth('obj:modify')
     def ensure_ip(self, spec=None, **kwarg):
         spec = spec or kwarg
 
@@ -776,7 +743,6 @@ class Interface(RTNL_Object):
         self._apply_script.append((do_ensure_ip, {'spec': spec}))
         return self
 
-    @check_auth('obj:modify')
     def del_ip(self, spec=None, **kwarg):
         spec = spec or kwarg
 
@@ -787,9 +753,11 @@ class Interface(RTNL_Object):
             elif callable(spec):
                 specs = self.ipaddr.dump()
                 specs.select_records(spec)
+                specs.materialize()
             else:
                 specs = self.ipaddr.dump()
                 specs.select_records(**spec)
+                specs.materialize()
             for sp in specs:
                 try:
                     method = getattr(self.ipaddr.locate(sp).remove(), mode)
@@ -800,22 +768,22 @@ class Interface(RTNL_Object):
                     e_s.trace = traceback.format_stack()
                     ret.append(e_s)
             if not ret:
-                ret = KeyError('no address records matched')
+                ret = [KeyError('no address records matched')]
             return ret
 
         self._apply_script.append((do_del_ip, {'spec': spec}))
         return self
 
-    @check_auth('obj:modify')
-    def add_port(self, spec):
+    def add_port(self, spec=None, **kwarg):
+        spec = spec or kwarg
+
         def do_add_port(self, mode, spec):
             try:
                 port = self.view[spec]
                 if port['target'] != self['target']:
                     raise ValueError('target must be the same')
                 port['master'] = self['index']
-                getattr(port, mode)()
-                return [port]
+                return [getattr(port, mode)()]
             except Exception as e_s:
                 e_s.trace = traceback.format_stack()
                 return [e_s]
@@ -823,8 +791,9 @@ class Interface(RTNL_Object):
         self._apply_script.append((do_add_port, {'spec': spec}))
         return self
 
-    @check_auth('obj:modify')
-    def del_port(self, spec):
+    def del_port(self, spec=None, **kwarg):
+        spec = spec or kwarg
+
         def do_del_port(self, mode, spec):
             try:
                 port = self.view[spec]
@@ -833,8 +802,7 @@ class Interface(RTNL_Object):
                 if port['target'] != self['target']:
                     raise ValueError('target must be the same')
                 port['master'] = 0
-                getattr(port, mode)()
-                return [port]
+                return [getattr(port, mode)()]
             except Exception as e_s:
                 e_s.trace = traceback.format_stack()
                 return [e_s]
@@ -842,22 +810,27 @@ class Interface(RTNL_Object):
         self._apply_script.append((do_del_port, {'spec': spec}))
         return self
 
-    @check_auth('obj:modify')
     def add_altname(self, ifname):
         new_list = set(self['alt_ifname_list'])
         new_list.add(ifname)
         self['alt_ifname_list'] = list(new_list)
+        return self
 
-    @check_auth('obj:modify')
     def del_altname(self, ifname):
         new_list = set(self['alt_ifname_list'])
         new_list.remove(ifname)
         self['alt_ifname_list'] = list(new_list)
+        return self
 
-    @check_auth('obj:modify')
     def __setitem__(self, key, value):
         if key == 'peer':
             dict.__setitem__(self, key, value)
+        elif key == 'ifname':
+            if value in self['alt_ifname_list']:
+                self.del_altname(value)
+            if key in self and self.old_ifname is None:
+                self.old_ifname = self[key]
+            super(Interface, self).__setitem__(key, value)
         elif key == 'target' and self.state == 'invalid':
             dict.__setitem__(self, key, value)
         elif key == 'net_ns_fd' and self.state == 'invalid':
@@ -870,7 +843,7 @@ class Interface(RTNL_Object):
             super(Interface, self).__setitem__(key, value)
 
     @classmethod
-    def spec_normalize(cls, processed, spec):
+    def spec_normalize(cls, spec):
         '''
         Interface key normalization::
 
@@ -880,10 +853,15 @@ class Interface(RTNL_Object):
 
         '''
         if isinstance(spec, basestring):
-            processed['ifname'] = spec
+            return {'ifname': spec}
         elif isinstance(spec, int):
-            processed['index'] = spec
-        return processed
+            return {'index': spec}
+        elif isinstance(spec, ifinfmsg):
+            ret = {'ifname': spec.get('ifname')}
+            if spec.get('index') != 0:
+                ret['index'] = spec.get('index')
+            return ret
+        return spec
 
     def complete_key(self, key):
         if isinstance(key, dict):
@@ -896,7 +874,7 @@ class Interface(RTNL_Object):
             ret_key['index'] = key
         return super(Interface, self).complete_key(ret_key)
 
-    def is_peer(self, other):
+    async def is_peer(self, other):
         '''Evaluate whether the given interface "points at" this one.'''
         if other['kind'] == 'vlan':
             return (
@@ -918,11 +896,10 @@ class Interface(RTNL_Object):
 
             other_link_netnsid = other.get('link_netnsid')
             if other_link_netnsid is not None:
-                self_source = self.sources[self['target']]
                 other_source = other.sources[other['target']]
-                info = other_source.api(
+                info = await other_source.api(
                     'get_netnsid',
-                    pid=self_source.api('get_pid'),
+                    pid=os.getpid(),
                     target_nsid=other_link_netnsid,
                 )
                 return info['current_nsid'] == other_link_netnsid
@@ -934,25 +911,23 @@ class Interface(RTNL_Object):
             'link', 'set', index=self['index'], xdp_fd=fd
         )
 
-    def snapshot(self, ctxid=None):
+    async def snapshot(self, ctxid=None):
         # 1. make own snapshot
-        snp = super(Interface, self).snapshot(ctxid=ctxid)
+        snp = await super().snapshot(ctxid=ctxid)
         # 2. collect dependencies and store in self.snapshot_deps
-        for spec in self.ndb.interfaces.getmany(
+        for spec in self.ndb.interfaces.asyncore.getmany(
             {'IFLA_MASTER': self['index']}
         ):
             # bridge ports
-            link = type(self)(
-                self.view, spec, auth_managers=self.auth_managers
-            )
-            snp.snapshot_deps.append((link, link.snapshot()))
-        for spec in self.ndb.interfaces.getmany({'IFLA_LINK': self['index']}):
-            link = type(self)(
-                self.view, spec, auth_managers=self.auth_managers
-            )
+            link = type(self)(self.view, spec)
+            snp.snapshot_deps.append((link, await link.snapshot()))
+        for spec in self.ndb.interfaces.asyncore.getmany(
+            {'IFLA_LINK': self['index']}
+        ):
+            link = type(self)(self.view, spec)
             # vlans & veth
-            if self.is_peer(link) and not link.is_peer(self):
-                snp.snapshot_deps.append((link, link.snapshot()))
+            if await self.is_peer(link) and not await link.is_peer(self):
+                snp.snapshot_deps.append((link, await link.snapshot()))
         # return the root node
         return snp
 
@@ -976,44 +951,79 @@ class Interface(RTNL_Object):
                         req[key] = self[key]
         return req
 
-    @check_auth('obj:modify')
-    def apply_altnames(self, alt_ifname_setup):
-        alt_ifname_remove = set(self['alt_ifname_list']) - alt_ifname_setup
-        alt_ifname_add = alt_ifname_setup - set(self['alt_ifname_list'])
+    async def apply_altnames(
+        self, alt_ifname_setup, alt_ifname_current, old_ifname=None
+    ):
+        if 'alt_ifname_list' in self.changed:
+            self.changed.remove('alt_ifname_list')
+        if alt_ifname_current is None:
+            # load the current state
+            await self.load_from_system()
+            self.load_sql(set_state=False)
+            alt_ifname_current = set(self['alt_ifname_list'])
+
+        alt_ifname_remove = alt_ifname_current - alt_ifname_setup
+        alt_ifname_add = alt_ifname_setup - alt_ifname_current
         for ifname in alt_ifname_remove:
-            self.sources[self['target']].api(
+            await self.sources[self['target']].api(
                 'link', 'property_del', index=self['index'], altname=ifname
             )
         for ifname in alt_ifname_add:
-            self.sources[self['target']].api(
+            await self.sources[self['target']].api(
                 'link', 'property_add', index=self['index'], altname=ifname
             )
-        self.load_from_system()
+        # reload alt ifnames from the system to check the state
+        await self.load_from_system()
         self.load_sql(set_state=False)
+        if old_ifname is not None and old_ifname in self['alt_ifname_list']:
+            alt_ifname_setup.add(old_ifname)
         if set(self['alt_ifname_list']) != alt_ifname_setup:
             raise Exception('could not setup alt ifnames')
 
-    @check_auth('obj:modify')
-    def apply(self, rollback=False, req_filter=None, mode='apply'):
+    async def apply(self, rollback=False, req_filter=None, mode='apply'):
         # translate string link references into numbers
         for key in ('link', 'master'):
             if key in self and isinstance(self[key], basestring):
-                self[key] = self.ndb.interfaces[self[key]]['index']
+                self[key] = self.ndb.interfaces.asyncore[self[key]]['index']
         setns = self.state.get() == 'setns'
         remove = self.state.get() == 'remove'
         alt_ifname_setup = set(self['alt_ifname_list'])
-        if 'alt_ifname_list' in self.changed:
-            self.changed.remove('alt_ifname_list')
+        old_ifname = self.old_ifname if 'ifname' in self.changed else None
         try:
-            super(Interface, self).apply(rollback, req_filter, mode)
-            if setns:
+            if 'index' in self and (
+                self.old_ifname or 'alt_ifname_list' in self.changed
+            ):
+                await self.apply_altnames(alt_ifname_setup, None)
+            if 'alt_ifname_list' in self.changed:
+                self.changed.remove('alt_ifname_list')
+            await super().apply(rollback, req_filter, mode)
+            if setns and self['net_ns_fd'] in self.sources:
                 self.load_value('target', self['net_ns_fd'])
                 dict.__setitem__(self, 'net_ns_fd', None)
+                for link in await self.sources[self['target']].api(
+                    'link', 'get', ifname=self['ifname']
+                ):
+                    # after interface move the name is the same,
+                    # but the index may change
+                    #
+                    # in this case .load_sql() will not update
+                    # the object, and the engine will try to apply
+                    # the interface's attributes to another interface
+                    # with the same index as our old one
+                    #
+                    # so resync the index first
+                    #
+                    # Bug-Url: https://github.com/svinota/pyroute2/issues/1181
+                    #
+                    self.load_value('index', link['index'])
+                    break
                 spec = self.load_sql()
                 if spec:
                     self.state.set('system')
-            if not remove:
-                self.apply_altnames(alt_ifname_setup)
+            if not remove and self.state != 'invalid':
+                await self.apply_altnames(
+                    alt_ifname_setup, set(self['alt_ifname_list']), old_ifname
+                )
 
         except NetlinkError as e:
             if (
@@ -1035,8 +1045,8 @@ class Interface(RTNL_Object):
                         ]
                     )
 
-                self.apply(rollback, req_filter, mode)
-                self.apply(rollback, None, mode)
+                await self.apply(rollback, req_filter, mode)
+                await self.apply(rollback, None, mode)
 
             elif (
                 e.code == 95
@@ -1056,22 +1066,24 @@ class Interface(RTNL_Object):
                         ]
                     )
 
-                self.apply(rollback, req_filter, mode)
+                await self.apply(rollback, req_filter, mode)
             else:
                 raise
+        finally:
+            self.old_ifname = None
         if ('net_ns_fd' in self.get('peer', {})) and (
             self['peer']['net_ns_fd'] in self.view.ndb.sources
         ):
             # wait for the peer in net_ns_fd, only if the netns
             # is connected to the NDB instance
-            self.view.wait(
+            await self.view.wait(
                 target=self['peer']['net_ns_fd'],
                 ifname=self['peer']['ifname'],
                 timeout=5,
             )
         return self
 
-    def hook_apply(self, method, **spec):
+    async def hook_apply(self, method, **spec):
         if method == 'set':
             if self['kind'] == 'bridge':
                 keys = filter(lambda x: x.startswith('br_'), self.changed)
@@ -1083,52 +1095,119 @@ class Interface(RTNL_Object):
                     }
                     for key in keys:
                         req[key] = self[key]
-                    self.sources[self['target']].api(self.api, method, **req)
+                    await self.sources[self['target']].api(
+                        self.api, method, **req
+                    )
                     # FIXME: make a reasonable shortcut for this
-                    self.load_from_system()
+                    await self.load_from_system()
             elif self['kind'] in ip_tunnels and self['state'] == 'down':
                 # force reading attributes for tunnels in the down state
-                self.load_from_system()
+                await self.load_from_system()
         elif method == 'add':
             if self['kind'] == 'tun':
                 self.load_sql()
-                self.load_event.wait(0.1)
+                await self.load_event.wait()
                 if 'index' not in self:
                     raise NetlinkError(errno.EAGAIN)
-                update = self.sources[self['target']].api(
+                update = await self.sources[self['target']].api(
                     self.api, 'get', index=self['index']
                 )
                 self.ndb._event_queue.put(update)
 
-    def load_from_system(self):
-        (
-            self.ndb._event_queue.put(
-                self.sources[self['target']].api(
-                    self.api, 'get', index=self['index']
-                )
-            )
+    async def load_from_system(self):
+        self.load_event.clear()
+        await self.sources[self['target']].api(
+            self.api, 'get', index=self['index']
         )
+        await self.load_event.wait()
 
     def load_sql(self, *argv, **kwarg):
         spec = super(Interface, self).load_sql(*argv, **kwarg)
         if spec:
-            tname = 'ifinfo_%s' % self['kind']
+            tname = f'ifinfo_{self["kind"]}'
             if tname in self.schema.compiled:
                 names = self.schema.compiled[tname]['norm_names']
-                spec = self.ndb.task_manager.db_fetchone(
-                    'SELECT * from %s WHERE f_index = %s'
-                    % (tname, self.schema.plch),
+                spec = self.ndb.schema.fetchone(
+                    f'SELECT * from {tname} WHERE f_index = ?',
                     (self['index'],),
                 )
                 if spec:
                     self.update(dict(zip(names, spec)))
         return spec
 
-    def load_rtnlmsg(self, *argv, **kwarg):
-        super(Interface, self).load_rtnlmsg(*argv, **kwarg)
+    async def load_rtnlmsg(self, *argv, **kwarg):
+        await super().load_rtnlmsg(*argv, **kwarg)
 
     def key_repr(self):
         return '%s/%s' % (
             self.get('target', ''),
             self.get('ifname', self.get('index', '')),
         )
+
+
+class SyncInterface(RTNL_Object):
+
+    def __init__(self, event_loop, obj, class_map=None, flags=Flags.RO):
+        super().__init__(event_loop, obj, class_map, flags)
+        self.ipaddr = SyncView(
+            event_loop, obj.ipaddr, self.class_map, self.flags
+        )
+        self.neighbours = SyncView(
+            event_loop, obj.neighbours, self.class_map, self.flags
+        )
+        self.ports = SyncView(
+            event_loop, obj.ports, self.class_map, self.flags
+        )
+        self.routes = SyncView(
+            event_loop, obj.routes, self.class_map, self.flags
+        )
+        self.vlans = SyncView(
+            event_loop, obj.vlans, self.class_map, self.flags
+        )
+
+    @property
+    def state(self):
+        return self.asyncore.state
+
+    def add_ip(self, spec=None, **kwarg):
+        self._main_sync_call(self.asyncore.add_ip, spec, **kwarg)
+        return self
+
+    def del_ip(self, spec=None, **kwarg):
+        self._main_sync_call(self.asyncore.del_ip, spec, **kwarg)
+        return self
+
+    def ensure_ip(self, spec=None, **kwarg):
+        self._main_sync_call(self.asyncore.ensure_ip, spec, **kwarg)
+        return self
+
+    def add_neighbour(self, spec=None, **kwarg):
+        self._main_sync_call(self.asyncore.add_neighbour, spec, **kwarg)
+        return self
+
+    def del_neighbour(self, spec=None, **kwarg):
+        self._main_sync_call(self.asyncore.del_neighbour, spec, **kwarg)
+        return self
+
+    def ensure_neighbour(self, spec=None, **kwarg):
+        self._main_sync_call(self.asyncore.ensure_neighbour, spec, **kwarg)
+        return self
+
+    def add_port(self, spec=None, **kwarg):
+        self._main_sync_call(self.asyncore.add_port, spec, **kwarg)
+        return self
+
+    def del_port(self, spec=None, **kwarg):
+        self._main_sync_call(self.asyncore.del_port, spec, **kwarg)
+        return self
+
+    def add_altname(self, ifname):
+        self._main_sync_call(self.asyncore.add_altname, ifname)
+        return self
+
+    def del_altname(self, ifname):
+        self._main_sync_call(self.asyncore.del_altname, ifname)
+        return self
+
+    def load_from_system(self):
+        self._main_async_call(self.asyncore.load_from_system)
