@@ -1,6 +1,13 @@
-from socket import AF_INET6
+from socket import AF_INET, AF_INET6
 
 from pyroute2.common import AF_MPLS
+from pyroute2.netlink.rt_files import (
+    RtDsfieldFile,
+    RtProtosFile,
+    RtRealmsFile,
+    RtScopesFile,
+    RtTablesFile,
+)
 from pyroute2.netlink.rtnl import encap_type, rt_proto, rt_scope, rt_type
 from pyroute2.netlink.rtnl.rtmsg import IP6_RT_PRIO_USER, LWTUNNEL_ENCAP_MPLS
 from pyroute2.netlink.rtnl.rtmsg import nh as nh_header
@@ -78,12 +85,20 @@ class RouteFieldFilter(IPTargets, NLAKeyTransform):
 
     def set_scope(self, context, value):
         if isinstance(value, str):
-            return {'scope': rt_scope[value]}
+            try:
+                value = rt_scope[value]
+            except KeyError:
+                # lookup with rt_scopes* files
+                value = RtScopesFile().get_rt_id(value)
         return {'scope': value}
 
     def set_proto(self, context, value):
         if isinstance(value, str):
-            return {'proto': rt_proto[value]}
+            try:
+                value = rt_proto[value]
+            except KeyError:
+                # lookup with rt_proto* files
+                value = RtProtosFile().get_rt_id(value)
         return {'proto': value}
 
     def set_encap_type(self, context, value):
@@ -96,8 +111,34 @@ class RouteFieldFilter(IPTargets, NLAKeyTransform):
             return {'type': rt_type[value]}
         return {'type': value}
 
+    def set_table(self, context, value):
+        if isinstance(value, str):
+            value = RtTablesFile().get_rt_id(value, 0)
+        return {"table": value}
+
+    def set_tos(self, context, value):
+        if isinstance(value, str):
+            value = RtDsfieldFile().get_rt_id(value, 0)
+        return {"tos": value}
+
 
 class RouteIPRouteFilter(IPRouteFilter):
+
+    def get_table(self, context, mode):
+        table = context.get('table', 0)
+        if mode == 'field':
+            if context.parameters['strict_check']:
+                return 0
+            return table if 0 < table < 255 else 254
+        return table
+
+    def set_flow(self, context, value):
+        if isinstance(value, str):
+            value = RtRealmsFile().get_rt_id(value)
+        return {"flow": value}
+
+    set_realms = set_flow
+
     def set_metrics(self, context, value):
         if value and 'attrs' not in value:
             metrics = {'attrs': []}
@@ -110,11 +151,12 @@ class RouteIPRouteFilter(IPRouteFilter):
         return {}
 
     def set_multipath(self, context, value):
+        ret = {}
         if value:
-            ret = []
+            hops = []
             for v in value:
                 if 'attrs' in v:
-                    ret.append(v)
+                    hops.append(v)
                     continue
                 nh = {'attrs': []}
                 nh_fields = [x[0] for x in nh_header.fields]
@@ -147,10 +189,20 @@ class RouteIPRouteFilter(IPRouteFilter):
                     else:
                         rta = rtmsg.name2nla(name)
                         nh['attrs'].append([rta, v[name]])
-                ret.append(nh)
-            if ret:
-                return {'multipath': ret}
-        return {}
+                hops.append(nh)
+            if hops:
+                ret = {'multipath': hops}
+                if context.get('family') is None:
+                    # autodetect and propagate family
+                    hop = ret[0]
+                    attrs = hop.get('attrs', [])
+                    for attr in attrs:
+                        if attr[0] == 'RTA_GATEWAY':
+                            ret['family'] = (
+                                AF_INET6 if attr[1].find(':') >= 0 else AF_INET
+                            )
+                            break
+        return ret
 
     def set_encap(self, context, value):
         if (
@@ -246,6 +298,21 @@ class RouteIPRouteFilter(IPRouteFilter):
         return {}
 
     def finalize(self, context):
+        # cleanup extra NLA for AF_MPLS
+        if context.get('family') == AF_MPLS:
+            for key in tuple(context.keys()):
+                if key not in (
+                    'family',
+                    'proto',
+                    'type',
+                    'dst',
+                    'newdst',
+                    'via',
+                    'multipath',
+                    'oif',
+                ):
+                    context.pop(key, None)
+        # cleanup empty strings
         for key in context:
             if context[key] in ('', None):
                 try:
