@@ -8,21 +8,30 @@ It helps to add/remove entries, list content, etc.
 For example, adding an entry with :class:`pyroute2.ipset.IPSet` object
 implies to set a various number of parameters:
 
->>> ipset = IPSet()
->>> ipset.add("foo", "1.2.3.4/24", etype="net")
->>> ipset.close()
+.. doctest::
+    :skipif: True
+
+    >>> ipset = IPSet()
+    >>> ipset.add("foo", "1.2.3.4/24", etype="net")
+    >>> ipset.close()
 
 When they are discovered by a :class:`WiSet`:
 
->>> wiset = load_ipset("foo")
->>> wiset.add("1.2.3.4/24")
+.. doctest::
+    :skipif: True
+
+    >>> wiset = load_ipset("foo")
+    >>> wiset.add("1.2.3.4/24")
 
 Listing entries is also easier using :class:`WiSet`, since it parses for you
 netlink messages:
 
->>> wiset.content
-{'1.2.3.0/24': IPStats(packets=None, bytes=None, comment=None,
-                       timeout=None, skbmark=None, physdev=False)}
+.. doctest::
+    :skipif: True
+
+    >>> wiset.content
+    {'1.2.3.0/24': IPStats(packets=None, bytes=None, comment=None,
+                           timeout=None, skbmark=None, physdev=False)}
 '''
 
 import errno
@@ -31,23 +40,24 @@ from collections import namedtuple
 from inspect import getcallargs
 from socket import AF_INET
 
-from pyroute2 import IPSet
 from pyroute2.common import basestring
+from pyroute2.ipset import IPSet
 from pyroute2.netlink.exceptions import IPSetError
-from pyroute2.netlink.nfnetlink.ipset import IPSET_FLAG_WITH_COUNTERS
-from pyroute2.netlink.nfnetlink.ipset import IPSET_FLAG_WITH_COMMENT
-from pyroute2.netlink.nfnetlink.ipset import IPSET_FLAG_WITH_SKBINFO
-from pyroute2.netlink.nfnetlink.ipset import IPSET_FLAG_PHYSDEV
-from pyroute2.netlink.nfnetlink.ipset import IPSET_FLAG_IFACE_WILDCARD
+from pyroute2.netlink.nfnetlink.ipset import (
+    IPSET_FLAG_IFACE_WILDCARD,
+    IPSET_FLAG_PHYSDEV,
+    IPSET_FLAG_WITH_COMMENT,
+    IPSET_FLAG_WITH_COUNTERS,
+    IPSET_FLAG_WITH_SKBINFO,
+)
 from pyroute2.netlink.nfnetlink.nfctsocket import IP_PROTOCOLS
 
-
 # Debug variable to detect netlink socket leaks
 COUNT = {"count": 0}
 
 
 def need_ipset_socket(fun):
-    """ Decorator to create netlink socket if needed.
+    """Decorator to create netlink socket if needed.
 
     In many of our helpers, we need to open a netlink socket. This can
     be expensive for someone using many times the functions: instead to have
@@ -63,6 +73,7 @@ def need_ipset_socket(fun):
     Note that all functions using this helper *must* use ipset as variable
     name for the socket.
     """
+
     def wrap(*args, **kwargs):
         callargs = getcallargs(fun, *args, **kwargs)
         if callargs["sock"] is None:
@@ -81,21 +92,47 @@ def need_ipset_socket(fun):
     return wrap
 
 
-class IPStats(namedtuple("IPStats", ["packets", "bytes", "comment",
-                                     "timeout", "skbmark", "physdev",
-                                     "wildcard"])):
+class IPStats(
+    namedtuple(
+        "IPStats",
+        [
+            "packets",
+            "bytes",
+            "comment",
+            "timeout",
+            "skbmark",
+            "physdev",
+            "wildcard",
+        ],
+    )
+):
     __slots__ = ()
 
-    def __new__(cls, packets, bytes, comment, timeout, skbmark, physdev=False,
-                wildcard=False):
-        return super(IPStats, cls).__new__(cls, packets, bytes, comment,
-                                           timeout, skbmark, physdev=physdev,
-                                           wildcard=wildcard)
+    def __new__(
+        cls,
+        packets,
+        bytes,
+        comment,
+        timeout,
+        skbmark,
+        physdev=False,
+        wildcard=False,
+    ):
+        return super(IPStats, cls).__new__(
+            cls,
+            packets,
+            bytes,
+            comment,
+            timeout,
+            skbmark,
+            physdev=physdev,
+            wildcard=wildcard,
+        )
 
 
 # pylint: disable=too-many-instance-attributes
 class WiSet(object):
-    """ Main high level ipset manipulation class.
+    """Main high level ipset manipulation class.
 
     Every high level ipset operation should be possible with this class,
     you probably don't need other helpers of this module, except tools
@@ -103,9 +140,12 @@ class WiSet(object):
 
     For example, you can create and an entry in a ipset just with:
 
-    >>> with WiSet(name="mysuperipset") as myset:
-    >>>    myset.create()             # add the ipset in the kernel
-    >>>    myset.add("198.51.100.1")  # add one IP to the set
+    .. doctest::
+        :skipif: True
+
+        >>> with WiSet(name="mysuperipset") as myset:
+        >>>    myset.create()             # add the ipset in the kernel
+        >>>    myset.add("198.51.100.1")  # add one IP to the set
 
     Netlink sockets are opened by __enter__ and __exit__ function, so you don't
     have to manage it manually if you use the "with" keyword.
@@ -113,16 +153,22 @@ class WiSet(object):
     If you want to manage it manually (for example for long operation in
     a daemon), you can do the following:
 
-    >>> myset = WiSet(name="mysuperipset")
-    >>> myset.open_netlink()
-    >>> # do stuff
-    >>> myset.close_netlink()
+    .. doctest::
+        :skipif: True
+
+        >>> myset = WiSet(name="mysuperipset")
+        >>> myset.open_netlink()
+        >>> # do stuff
+        >>> myset.close_netlink()
 
     You can also don't initiate at all any netlink socket, this code will work:
 
-    >>> myset = WiSet(name="mysuperipset")
-    >>> myset.create()
-    >>> myset.destroy()
+    .. doctest::
+        :skipif: True
+
+        >>> myset = WiSet(name="mysuperipset")
+        >>> myset.create()
+        >>> myset.destroy()
 
     But do it very carefully. In that case, a netlink socket will be opened
     in background for any operation. No socket will be leaked, but that
@@ -131,16 +177,29 @@ class WiSet(object):
     You can also instantiate WiSet objects with :func:`load_all_ipsets` and
     :func:`load_ipset`:
 
-    >>> all_sets_dict = load_all_ipsets()
-    >>> one_set = load_ipset(name="myset")
+    .. doctest::
+        :skipif: True
+
+        >>> all_sets_dict = load_all_ipsets()
+        >>> one_set = load_ipset(name="myset")
 
     Have a look on content variable if you need list of entries in the Set.
     """
 
     # pylint: disable=too-many-arguments
-    def __init__(self, name=None, attr_type='hash:ip', family=AF_INET,
-                 sock=None, timeout=None, counters=False, comment=False,
-                 hashsize=None, revision=None, skbinfo=False):
+    def __init__(
+        self,
+        name=None,
+        attr_type='hash:ip',
+        family=AF_INET,
+        sock=None,
+        timeout=None,
+        counters=False,
+        comment=False,
+        hashsize=None,
+        revision=None,
+        skbinfo=False,
+    ):
         self.name = name
         self.hashsize = hashsize
         self._attr_type = None
@@ -157,13 +216,16 @@ class WiSet(object):
         self.skbinfo = skbinfo
 
     def open_netlink(self):
-        """ Open manually a netlink socket. You can use "with WiSet()" instead
+        """
+        Open manually a netlink socket.
+
+        You can use "with WiSet()" statement instead.
         """
         if self.sock is None:
             self.sock = IPSet()
 
     def close_netlink(self):
-        """ Clone any opened netlink socket """
+        """Clone any opened netlink socket"""
         if self.sock is not None:
             self.sock.close()
             self.sock = None
@@ -186,7 +248,7 @@ class WiSet(object):
 
     @classmethod
     def from_netlink(cls, ndmsg, content=False):
-        """ Create a ipset objects based on a parsed netlink message
+        """Create a ipset objects based on a parsed netlink message
 
         :param ndmsg: the netlink message to parse
         :param content: should we fill (and parse) entries info (can be slow
@@ -214,7 +276,7 @@ class WiSet(object):
         return self
 
     def update_dict_content(self, ndmsg):
-        """ Update a dictionary statistics with values sent in netlink message
+        """Update a dictionary statistics with values sent in netlink message
 
         :param ndmsg: the netlink message
         :type ndmsg: netlink message
@@ -251,6 +313,8 @@ class WiSet(object):
                         proto = IP_PROTOCOLS.get(proto, str(proto)).lower()
                         key += '{proto}:'.format(proto=proto)
                     key += str(entry.get_attr("IPSET_ATTR_PORT_FROM"))
+                elif parse_type == "mac":
+                    key += entry.get_attr("IPSET_ATTR_ETHER")
                 key += ","
 
             key = key.strip(",")
@@ -268,33 +332,43 @@ class WiSet(object):
             entry_flag_parsed = {"physdev": False}
             flags = entry.get_attr("IPSET_ATTR_CADT_FLAGS")
             if flags is not None:
-                entry_flag_parsed["physdev"] = bool(flags &
-                                                    IPSET_FLAG_PHYSDEV)
-                entry_flag_parsed["wildcard"] = bool(flags &
-                                                     IPSET_FLAG_IFACE_WILDCARD)
+                entry_flag_parsed["physdev"] = bool(flags & IPSET_FLAG_PHYSDEV)
+                entry_flag_parsed["wildcard"] = bool(
+                    flags & IPSET_FLAG_IFACE_WILDCARD
+                )
 
-            value = IPStats(packets=entry.get_attr("IPSET_ATTR_PACKETS"),
-                            bytes=entry.get_attr("IPSET_ATTR_BYTES"),
-                            comment=entry.get_attr("IPSET_ATTR_COMMENT"),
-                            skbmark=skbmark,
-                            timeout=timeout,
-                            **entry_flag_parsed)
+            value = IPStats(
+                packets=entry.get_attr("IPSET_ATTR_PACKETS"),
+                bytes=entry.get_attr("IPSET_ATTR_BYTES"),
+                comment=entry.get_attr("IPSET_ATTR_COMMENT"),
+                skbmark=skbmark,
+                timeout=timeout,
+                **entry_flag_parsed,
+            )
             self._content[key] = value
 
     def create(self, **kwargs):
-        """ Insert this Set in the kernel
+        """Insert this Set in the kernel
 
         Many options are set with python object attributes (like comments,
         counters, etc). For non-supported type, kwargs are provided. See
         :doc:`ipset` documentation for more information.
         """
-        create_ipset(self.name, stype=self.attr_type, family=self.family,
-                     sock=self.sock, timeout=self.timeout,
-                     comment=self.comment, counters=self.counters,
-                     hashsize=self.hashsize, skbinfo=self.skbinfo, **kwargs)
+        create_ipset(
+            self.name,
+            stype=self.attr_type,
+            family=self.family,
+            sock=self.sock,
+            timeout=self.timeout,
+            comment=self.comment,
+            counters=self.counters,
+            hashsize=self.hashsize,
+            skbinfo=self.skbinfo,
+            **kwargs,
+        )
 
     def destroy(self):
-        """ Destroy this ipset in the kernel list.
+        """Destroy this ipset in the kernel list.
 
         It does not delete this python object (any content or other stored
         values are keep in memory). This function will fail if the ipset is
@@ -303,7 +377,7 @@ class WiSet(object):
         destroy_ipset(self.name, sock=self.sock)
 
     def add(self, entry, **kwargs):
-        """ Add an entry in this ipset.
+        """Add an entry in this ipset.
 
         If counters are enabled on the set, reset by default the value when
         we add the element. Without this reset, kernel sometimes store old
@@ -324,39 +398,43 @@ class WiSet(object):
             except IndexError:
                 mask = int("0xffffffff", 16)
             kwargs["skbmark"] = (mark, mask)
-        add_ipset_entry(self.name, entry, etype=self.entry_type,
-                        sock=self.sock, **kwargs)
+        add_ipset_entry(
+            self.name, entry, etype=self.entry_type, sock=self.sock, **kwargs
+        )
 
     def delete(self, entry, **kwargs):
-        """ Delete/remove an entry in this ipset """
-        delete_ipset_entry(self.name, entry, etype=self.entry_type,
-                           sock=self.sock, **kwargs)
+        """Delete/remove an entry in this ipset"""
+        delete_ipset_entry(
+            self.name, entry, etype=self.entry_type, sock=self.sock, **kwargs
+        )
 
     def test(self, entry, **kwargs):
-        """ Test if an entry is in this ipset """
-        return test_ipset_entry(self.name, entry, etype=self.entry_type,
-                                sock=self.sock, **kwargs)
+        """Test if an entry is in this ipset"""
+        return test_ipset_entry(
+            self.name, entry, etype=self.entry_type, sock=self.sock, **kwargs
+        )
 
     def test_list(self, entries, **kwargs):
-        """ Test if a list of a set of entries is in this ipset
+        """Test if a list of a set of entries is in this ipset
 
         Return a set of entries found in the IPSet
         """
-        return test_ipset_entries(self.name, entries, etype=self.entry_type,
-                                  sock=self.sock, **kwargs)
+        return test_ipset_entries(
+            self.name, entries, etype=self.entry_type, sock=self.sock, **kwargs
+        )
 
     def update_content(self):
-        """ Update the content dictionary with values from kernel """
+        """Update the content dictionary with values from kernel"""
         self._content = {}
         update_wiset_content(self, sock=self.sock)
 
     def flush(self):
-        """ Flush entries of the ipset """
+        """Flush entries of the ipset"""
         flush_ipset(self.name, sock=self.sock)
 
     @property
     def content(self):
-        """ Dictionary of entries in the set.
+        """Dictionary of entries in the set.
 
         Keys are IP addresses (as string), values are IPStats tuples.
         """
@@ -366,19 +444,19 @@ class WiSet(object):
         return self._content
 
     def insert_list(self, entries):
-        """ Just a small helper to reduce the number of loops in main code. """
+        """Just a small helper to reduce the number of loops in main code."""
         for entry in entries:
             self.add(entry)
 
     def replace_entries(self, new_list):
-        """ Replace the content of an ipset with a new list of entries.
+        """Replace the content of an ipset with a new list of entries.
 
         This operation is like a flush() and adding all entries one by one. But
         this call is atomic: it creates a temporary ipset and swap the content.
 
         :param new_list: list of entries to add
         :type new_list: list or :py:class:`set` of basestring or of
-        keyword arguments dict
+            keyword arguments dict
         """
         temp_name = str(uuid.uuid4())[0:8]
         # Get a copy of ourself
@@ -392,16 +470,18 @@ class WiSet(object):
 
 
 @need_ipset_socket
-def create_ipset(name, stype=None, family=AF_INET, exclusive=False,
-                 sock=None, **kwargs):
-    """ Create an ipset. """
-    sock.create(name, stype=stype, family=family, exclusive=exclusive,
-                **kwargs)
+def create_ipset(
+    name, stype=None, family=AF_INET, exclusive=False, sock=None, **kwargs
+):
+    """Create an ipset."""
+    sock.create(
+        name, stype=stype, family=family, exclusive=exclusive, **kwargs
+    )
 
 
 @need_ipset_socket
 def load_all_ipsets(content=False, sock=None, inherit_sock=False, prefix=None):
-    """ List all ipset as WiSet objects.
+    """List all ipset as WiSet objects.
 
     Get full ipset data from kernel and parse it in WiSet objects. Result is
     a dictionary with ipset names as keys, and WiSet objects as values.
@@ -434,7 +514,7 @@ def load_all_ipsets(content=False, sock=None, inherit_sock=False, prefix=None):
 
 @need_ipset_socket
 def load_ipset(name, content=False, sock=None, inherit_sock=False):
-    """ Get one ipset as WiSet object
+    """Get one ipset as WiSet object
 
     Helper to get current WiSet object. More efficient that
     :func:`load_all_ipsets` since the kernel does the filtering itself.
@@ -468,7 +548,7 @@ def load_ipset(name, content=False, sock=None, inherit_sock=False):
 
 @need_ipset_socket
 def update_wiset_content(wiset, sock=None):
-    """ Update content/statistics of a wiset.
+    """Update content/statistics of a wiset.
 
     You should never call yourself this function. It is only a helper to use
     the :func:`need_ipset_socket` decorator out of WiSet object.
@@ -479,25 +559,25 @@ def update_wiset_content(wiset, sock=None):
 
 @need_ipset_socket
 def destroy_ipset(name, sock=None):
-    """ Remove an ipset in the kernel. """
+    """Remove an ipset in the kernel."""
     sock.destroy(name)
 
 
 @need_ipset_socket
 def add_ipset_entry(name, entry, sock=None, **kwargs):
-    """ Add an entry """
+    """Add an entry"""
     sock.add(name, entry, **kwargs)
 
 
 @need_ipset_socket
 def delete_ipset_entry(name, entry, sock=None, **kwargs):
-    """ Remove one entry """
+    """Remove one entry"""
     sock.delete(name, entry, **kwargs)
 
 
 @need_ipset_socket
 def test_ipset_exist(name, sock=None):
-    """ Test if the given ipset exist """
+    """Test if the given ipset exist"""
     try:
         sock.headers(name)
         return True
@@ -509,14 +589,13 @@ def test_ipset_exist(name, sock=None):
 
 @need_ipset_socket
 def test_ipset_entry(name, entry, sock=None, **kwargs):
-    """ Test if an entry is in one ipset """
+    """Test if an entry is in one ipset"""
     return sock.test(name, entry, **kwargs)
 
 
 @need_ipset_socket
 def test_ipset_entries(name, entries, sock=None, **kwargs):
-    """ Test a list (or a set) of entries.
-    """
+    """Test a list (or a set) of entries."""
     res = set()
     for entry in entries:
         if sock.test(name, entry, **kwargs):
@@ -526,13 +605,13 @@ def test_ipset_entries(name, entries, sock=None, **kwargs):
 
 @need_ipset_socket
 def flush_ipset(name, sock=None):
-    """ Flush all ipset content """
+    """Flush all ipset content"""
     sock.flush(name)
 
 
 @need_ipset_socket
 def swap_ipsets(name_a, name_b, sock=None):
-    """ Swap the content of ipset a and b.
+    """Swap the content of ipset a and b.
 
     ipsets must have compatible content.
     """
@@ -540,5 +619,5 @@ def swap_ipsets(name_a, name_b, sock=None):
 
 
 def get_ipset_socket(**kwargs):
-    """ Get a socket that one can pass to several WiSet objects """
+    """Get a socket that one can pass to several WiSet objects"""
     return IPSet(**kwargs)
